from ast import (  # Import,; ImportFrom,; Expr,; alias,
    Add,
    And,
    Assign,
    Attribute,
    BinOp,
    BoolOp,
    Call,
    Compare,
    Constant,
    Eq,
    FunctionDef,
    GeneratorExp,
    Gt,
    GtE,
    If,
    Import,
    Is,
    IsNot,
    Lambda,
    List,
    Load,
    Lt,
    LtE,
    Module,
    Name,
    Not,
    NotEq,
    Num,
    Return,
    Store,
    Str,
    Tuple,
    UnaryOp,
    While,
    alias,
    arg,
    arguments,
    comprehension,
    fix_missing_locations,
    stmt,
)
from enum import Enum

OPERATORS = {
    "<": Lt,
    "<=": LtE,
    "=": Eq,
    "!=": NotEq,
    ">": Gt,
    ">=": GtE,
}

_extensions = {}


class Parser(object):
    def __init__(self, function):
        expression = function(None)
        if getattr(expression, "copy", None):
            expression = expression.copy()
        func_code = getattr(function, "__code__", None) or function.func_code
        self.file = {
            "lineno": func_code.co_firstlineno,
            "filename": func_code.co_filename,
            "col_offset": 0,
            "ctx": Load(),
        }
        self.file_store = dict(self.file, ctx=Store())
        self.expression = expression
        self.imports = set()
        body = self.build_expression(expression)
        if isinstance(body, list):
            pass
        elif not isinstance(body, stmt):
            body = [Return(value=body, **self.file)]
        else:
            body = [body]

        self.ast = Module(
            # body=self.imports + [
            body=[
                FunctionDef(
                    name=func_code.co_name,
                    args=arguments(
                        args=[arg(arg="self", annotation=None)],  # function.func_code.co_varnames
                        defaults=[],
                        vararg=None,
                        kwarg=None,
                        kwonlyargs=[],
                        kw_defaults=[],
                        posonlyargs=[],
                    ),
                    kwarg=[],
                    kw_defaults=[],
                    vararg=[],
                    kwonlyargs=[],
                    body=[
                        Import(names=[alias(name=imp, **self.file)], **self.file)
                        for imp in self.imports
                    ] + body,
                    decorator_list=[],
                ),
            ],
            type_ignores=[],
            **self.file,
        )
        # import pdb; pdb.set_trace()
        fix_missing_locations(self.ast)
        self.code = compile(self.ast, mode="exec", filename=self.file["filename"])

    def build_expression(self, expression):
        if expression is None:
            return Constant(value=None, **self.file)
        handler_name = "handle_{}".format(expression.__class__.__name__.lower())

        if handler_name in _extensions:
            return _extensions[handler_name](self, expression)

        return getattr(self, handler_name)(expression)

    def handle_case(self, case):
        # We can have N When expressions, and then (optionally) one that is the default.
        return self.handle_when(*case.get_source_expressions())

    def handle_when(self, when, *others):
        test, body = when.get_source_expressions()
        if others:
            if others[0].__class__.__name__ == "When":
                orelse = [self.handle_when(*others)]
            else:
                # Can we ever have other expressions after the default?
                orelse = [Return(value=self.build_expression(others[0]), **self.file)]
        else:
            orelse = []

        return If(
            test=self.build_expression(test),
            body=[Return(value=self.build_expression(body), **self.file)],
            orelse=orelse,
            **self.file,
        )

    def handle_bool(self, boolean):
        return Constant(value=boolean, **self.file)

    def handle_q(self, q):
        if not q.children:
            expr = Name(id="True", **self.file)
        elif len(q.children) == 1:
            expr = self._attr_lookup(*q.children[0])
        elif q.connector == "AND":
            expr = Call(
                func=Name(id="all", **self.file),
                args=[List(elts=[self._attr_lookup(k, v) for k, v in q.children], **self.file)],
                keywords=[],
                kwonlyargs=[],
                **self.file,
            )
        else:  # q.connector == 'OR'
            expr = Call(
                func=Name(id="any", **self.file),
                args=[List(elts=[self._attr_lookup(k, v) for k, v in q.children], **self.file)],
                keywords=[],
                kwonlyargs=[],
                **self.file,
            )

        if q.negated:
            return UnaryOp(op=Not(), operand=expr, **self.file)

        return expr

    def handle_concat(self, concat):
        return self.build_expression(*concat.get_source_expressions())

    def handle_concatpair(self, pair):
        a, b = pair.get_source_expressions()
        return BinOp(left=self.build_expression(a), op=Add(), right=self.build_expression(b), **self.file)

    def handle_f(self, f):
        # Do we need to use .attname here?
        # What about transforms/lookups?
        f.contains_aggregate = False
        if "__" in f.name:
            self.imports.add('functools')

            return Call(
                func=Attribute(value=Name(id='functools', **self.file), attr='reduce', **self.file),
                args=[
                    Lambda(
                        args=arguments(
                            posonlyargs=[],
                            args=[arg(arg='x', **self.file), arg(arg='y', **self.file)],
                            kwonlyargs=[],
                            kw_defaults=[],
                            defaults=[]
                        ),
                        body=Call(
                            func=Name(id='getattr', **self.file),
                            args=[
                                Name(id='x', **self.file),
                                Name(id='y', **self.file),
                                Constant(value=None, **self.file),
                            ],
                            keywords=[],
                            **self.file
                        ),
                        **self.file
                    ),
                    List(
                        elts=[
                            Constant(value=x, **self.file)
                            for x in f.name.split('__')
                        ],
                        **self.file
                    ),
                    Name(id='self', **self.file),
                ],
                keywords=[],
                **self.file,
            )

            left, *parts = f.name.split('__')
            expression = Attribute(value=Name(id='self', **self.file), attr=left, **self.file)
            while parts:
                left, *parts = parts
                expression = Attribute(value=expression, attr=left, **self.file)
            return expression

            # We need to chain a bunch of attr lookups, returning None
            # if any of them give us a None, we should be returning a None.
            #

            # .  while x is not None and parts:
            # .      x = getattr(x, parts.pop(0), None)
            #   return x
            return [
                Assign(
                    targets=[Name(id="source", **self.file_store)], value=Name(id="self", **self.file), **self.file
                ),
                Assign(
                    targets=[Name(id="parts", **self.file_store)],
                    value=Call(
                        func=Attribute(value=Constant(value=f.name, **self.file), attr="split", **self.file),
                        args=[Constant(value="__", **self.file)],
                        keywords=[],
                        kwonlyargs=[],
                        **self.file,
                    ),
                    **self.file,
                ),
                While(
                    test=BoolOp(
                        op=And(),
                        values=[
                            Compare(
                                left=Name(id="source", **self.file),
                                ops=[IsNot()],
                                comparators=[Constant(value=None, **self.file)],
                                **self.file,
                            ),
                            Name(id="parts", **self.file),
                        ],
                        **self.file,
                    ),
                    body=[
                        Assign(
                            targets=[Name(id="source", **self.file_store)],
                            value=Call(
                                func=Name(id="getattr", **self.file),
                                args=[
                                    Name(id="source", **self.file),
                                    Call(
                                        func=Attribute(value=Name(id="parts", **self.file), attr="pop", **self.file),
                                        args=[Constant(value=0, **self.file)],
                                        keywords=[],
                                        kwonlyargs=[],
                                        **self.file,
                                    ),
                                    Constant(value=None, **self.file),
                                ],
                                keywords=[],
                                kwonlyargs=[],
                                **self.file,
                            ),
                            **self.file,
                        ),
                    ],
                    orelse=[],
                    **self.file,
                ),
                Return(value=Name(id="source", **self.file), **self.file),
            ]
        return Attribute(value=Name(id="self", **self.file), attr=f.name, **self.file)

    def handle_value(self, value):
        if isinstance(value.value, Enum):
            return Attribute(
                value=Name(id=value.value.__class__.__name__, **self.file),
                attr=value.value.name,
                **self.file,
            )

        if isinstance(value.value, str):
            return Str(s=value.value, **self.file)

        if value.value is None:
            return Constant(value=None, **self.file)

        if isinstance(value.value, int):
            return Num(n=value.value, **self.file)

        if isinstance(value.value, bool):
            return Constant(value=value.value, **self.file)

        raise ValueError("Unhandled Value")

    def _handle_call_factory(func):
        def handle_call(self, expression):
            return Call(
                func=Attribute(
                    value=self.build_expression(*expression.get_source_expressions()), attr=func, **self.file
                ),
                args=[],
                keywords=[],
                kwonlyargs=[],
                **self.file,
            )

        return handle_call

    handle_upper = _handle_call_factory("upper")
    handle_lower = _handle_call_factory("lower")

    def handle_expressionwrapper(self, expression):
        return self.build_expression(*expression.get_source_expressions())

    def handle_combinedexpression(self, expression):
        return Compare(
            left=self.build_expression(expression.lhs),
            ops=[OPERATORS[expression.connector](**self.file)],
            comparators=[self.build_expression(expression.rhs)],
            **self.file,
        )

    # This is commented out because it only supports one function.
    # def handle_func(self, expression):
    #     func = expression.extra.get('function') or expression.function
    #     if func in {'current_timestamp', 'now'}:
    #         self.imports.append(Import(names=[alias(name='django.utils')]))
    #         return Call(
    #             func=Attribute(
    #                 value=Attribute(
    #                     value=Attribute(
    #                         value=Name('django', **self.file),
    #                         attr='utils',
    #                         **self.file
    #                     ),
    #                     attr='timezone',
    #                     **self.file
    #                 ),
    #                 attr='now',
    #                 **self.file
    #             ),
    #             args=[],
    #             keywords=[],
    #             **self.file,
    #         )
    #         return Call(func=Name(id='utcnow', **self.file), args=[], keywords=[], **self.file)

    def handle_coalesce(self, expression):
        """
        next(
            itertools.chain(
                (x for x in expression.get_source_expressions() where x is not None),
                (None,)
            )
        )
        """
        expressions = List(
            elts=[self.build_expression(exp) for exp in expression.get_source_expressions()],
            **self.file,
        )

        return Call(
            func=Name(id="next", **self.file),
            args=[
                GeneratorExp(
                    elt=Name(id="x", **self.file),
                    generators=[
                        comprehension(
                            target=Name(id="x", **(dict(self.file, ctx=Store()))),
                            iter=expressions,
                            ifs=[
                                Compare(
                                    left=Name(id="x", **self.file),
                                    ops=[IsNot()],
                                    comparators=[Constant(value=None, **self.file)],
                                ),
                            ],
                            is_async=False,
                            **self.file,
                        )
                    ],
                    **self.file,
                ),
                Tuple(elts=[Constant(value=None, **self.file)], **self.file),
            ],
            keywords=[],
            **self.file,
        )

    def handle_exact(self, exact):
        left, right = exact.get_source_expressions()
        return Compare(
            left=self.build_expression(left),
            ops=[Eq(**self.file)],
            comparators=[self.build_expression(right)],
            **self.file
        )

    def _attr_lookup(self, attr, value):
        if "__" not in attr:
            return Compare(
                left=Attribute(value=Name(id="self", **self.file), attr=attr, **self.file),
                ops=[Eq()],
                comparators=[self.build_expression(value)],
                **self.file,
            )

        attr, lookup = attr.split("__", 1)

        if lookup == "isnull":
            return Compare(
                left=Attribute(value=Name(id="self", **self.file), attr=attr, **self.file),
                ops=[Is() if value else IsNot()],
                comparators=[Constant(value=None, **self.file)],
                **self.file,
            )

        if lookup == "exact":
            return self._attr_lookup(attr, value)

        raise ValueError("Unhandled attr lookup")


class register:
    def __init__(self, func):
        if isinstance(func, str):
            self.name = f'handle_{func}'
        else:
            name = func.__code__.co_name
            if not name.startswith('handle_'):
                name = f'handle_{name}'
            _extensions[name] = func

    def __call__(self, func):
        _extensions[self.name] = func



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['django_shared_property']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'django-shared-property',
    'version': '0.5.3',
    'description': 'Properties that are both ORM expressions and python code.',
    'long_description': '======================\nDjango Shared Property\n======================\n\n\n.. image:: https://img.shields.io/pypi/v/django_shared_property.svg\n        :target: https://pypi.python.org/pypi/django_shared_property\n\n.. image:: https://img.shields.io/travis/schinckel/django-shared-property.svg\n        :target: https://travis-ci.org/schinckel/django-shared-property\n\n.. image:: https://ci.appveyor.com/api/projects/status/schinckel/branch/main?svg=true\n    :target: https://ci.appveyor.com/project/schinckel/django-shared-property/branch/main\n    :alt: Build status on Appveyor\n\n.. image:: https://readthedocs.org/projects/django-shared-property/badge/?version=latest\n        :target: https://django-shared-property.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\n\n\n\nProperties that are both ORM expressions and python code.\n\n\n* Free software: MIT license\n\n* Documentation: https://django-shared-property.readthedocs.io.\n\n\n\nInstallation:\n-------------\n\n.. code-block:: console\n\n    $ pip install django_shared_property\n\nPhilosophy:\n-------------\n\nI often find that I have annotations in Django querysets that are based on one or more other fields, and are used frequently. In some cases I have even been known to ensure these annotations are always available on the model using a custom queryset/manager.\n\nBut, unlike Python properties, these annotations are not "live". If, for example, you have the following:\n\n.. code-block:: python\n\n    class FullNameQueryset(models.query.QuerySet):\n        def with_full_name(self):\n            return self.annotate(\n              full_name=Concat(models.F(\'first_name\'), models.Value(\' \'), models.F(\'last_name\')),\n            )\n\n\n    class Person(models.Model):\n        first_name = models.TextField()\n        last_name = models.TextField()\n\n        objects = FullNameQueryset.as_manager()\n\n\nThen, you can do a `person = Person.objects.with_full_name().get(pk=1)`, and then reference `person.full_name`.\n\nBut, if you modify `person.first_name`, you\'d need to write it back to the database, and then reload it. Which may not be ideal, and at best requires two database operations.\n\ndjango-shared-property allows you to have properties that automatically act as an annototation, allowing you to define the expression, and have Django use that operation within any database query. It can then be used in a filter (or further annotation), even across relationships. And finally, if any local changes are made to the object that would affect the value when stored in the database, then the property value will also update in Python.\n\nShow me how!\n-------------\n\nSimilar to a Python property, a django-shared-property requires a method that takes no arguments. It should, however, return a Django Expression. For example, following our annotation above:\n\n.. code-block:: python\n\n    class Person(models.Model):\n        first_name = models.TextField()\n        last_name = models.TextField()\n\n        @shared_property\n        def full_name(self):\n            return Concat(models.F(\'first_name\'), models.Value(\' \'), models.F(\'last_name\'))\n\n\nYou then reference it just like any other field.\n\n.. code-block:: python\n\n\n    Person.objects.filter(full_name__contains=\'Bob\')\n\n\nWhat else can it do?\n---------------------\n\nShared properties can reference any number of fields on the model, and even other shared properties, just like with annotations. They can even reference fields from related models, using the familiar ``models.F(\'relation__field\')`` lookup syntax. You can also use some Django expressions (such as Concat, Lower, Upper), where there is a clear relationship to a Python concept.\n\n  * Case\n  * When\n  * F\n  * Q (Specifically, within a When, but it could work elsewhere)\n  * Concat\n  * Value\n  * Lower and Upper (but only on Python objects that have these as attributes)\n  * ExpressionWrapper\n  * CombinedExpresson\n  * Coalesce (but see the note below)\n\nWithin the context of a Q expression, you can use ``__isnull`` and ``__exact`` lookups.\n\n\nYou can even refer to constants in your Python file, such as the different values of an Enum. The return value of your python object will then correctly return instances of the Enum.\n\n\nIf your chosen expression/function/value does not work, then it may be possible to implement it (see below).\n\n\nShared properties should be pure functions - they must not refer to ``self`` (indeed, this will cause an error), and should not refer to variables, as they will be executed at times other than when they are about to be decorated.\n\n\nHow does it work?\n------------------\n\nBecause of the limit that the decorated function be a pure function, we are able to execute the callable, using the result as a Django expression.\n\nThe Django part is relatively straightforward. The expression returned by the method that is decorated as a shared_property is used in a context that looks a bit like an annotation - however there are a couple of things that need to be done to ensure that the expression has the correct data available to it to make sure it points at the correct tables. We indicate to Django  that it should not be written back to the database by marking it as a ``private`` field.\n\nCreating the Python property is a bit trickier. We still need the expression, but we build an Abstract Syntax Tree based on the expression. We then compile this into a callable object that we use as the property.\n\nIn a little more detail:\n\n  * Call the decorated function, returning the Expression\n  * Use Python\'s introspection tools to examine the expression (and it\'s "source expressions") and a recursive descent parser to build an AST equivalent to the expression. Specifically, the AST contains a function definition.\n  * Compile this AST into a code object\n  * ``eval`` this code object with the correct context to pull in any constants from outside the namespace.\n  * Extract the newly defined function, and use it for the callable in our property.\n\n\nAdvanced Use\n-------------\n\nSometimes you want to define the callable yourself: there is an alternate syntax for that. This could be where the expression has not been defined, or it\'s possible to create a more efficient callable by hand:\n\n.. code-block:: python\n\n    class MyModel(models.Model):\n        # other fields\n\n        @shared_property(Case(\n            When(models.Q(x__gte=2, x__lt=5), then=models.Value(\'B\')),\n            When(models.Q(x__lt=2), then=models.Value(\'A\')),\n            default=models.Value(\'C\'),\n            output_field=models.TextField(),\n        )\n        def state(self):\n            if 2 <= self.x < 5:\n                return \'B\'\n            elif x < 2:\n                return \'A\'\n            return \'C\'\n\n        @shared_property(Coalesce(\n          CombinedExpression(F(\'expiry_date\'), \'<\', Func(function=\'current_timestamp\')),\n          models.Value(True),\n        ))\n        def active(self):\n            return self.expiry_date is None or self.expiry_date < timezone.now()\n\n\nIn this specific case, the code that is generated would be fairly similar (although it would not use the ``a < b < c`` idiom), however it shows how it is possible to to explicitly provide the python code. Please note that the onus of responsibility is on the developer to ensure that the expression and function are equivalent in this context.\n\nThe second example shows where a python comparison doesn\'t quite map to the SQL code: the ``COALESCE(expiry_date < now(), true)`` relies on SQL comparisons involving NULL to also return NULL, but in Python you cannot do this.\n\nAlso note that in this case only a single expression may be used as the argument to the decorator.\n\nRegistering Expressions\n-----------------------\n\nIt is possible to register your own expressions. The structure is quite strict, and you\'ll need to reference the parser instance as well as the incoming expression. There\'s sometimes quite a bit of work to turn the Expression into (a) the correct Python, and then (b) the AST that is required.\n\n\n.. code-block:: python\n\n    from django_shared_property.parser import register\n\n\n    @register\n    def handle_foo(parser, expression):\n        # This assumes a foo() function in python that matches a foo()\n        # function in SQL, neither of which takes arguments.\n        return Call(\n            func=Name(id=\'foo\', **parser.file),\n            args=[],\n            keywords=[],\n            kwonlyargs=[],\n            **parser.file,\n        )\n\n\n    class Foo(Func):\n        function = \'foo\'\n\n\n    class MyModel(models.Model):\n        # ...\n\n        @shared_property\n        def the_foo(self):\n            return Foo()\n\n\nThis is a toy example - try looking in the ``parser`` module for other examples.\n\nLimitations\n-----------\n\nUse of the django queryset methods defer/only prevent any shared properties from loading. However, because of the way the feature works, you would still be able to use this property - at least in the case where the referenced fields are local.\n\nWhen you use a shared property that references a related model, and then try to filter on this, you cannot perform a count or exists query. See https://github.com/schinckel/django-shared-property/issues/2\n\n\nCredits\n-------\n\nDeveloped by `Matthew Schinckel`_.\n\nThis package was created with Cookiecutter_ and the `wboxx1/cookiecutter-pypackage-poetry`_ project template.\n\n.. _`Matthew Schinckel`: https://schinckel.net/\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`wboxx1/cookiecutter-pypackage-poetry`: https://github.com/wboxx1/cookiecutter-pypackage-poetry\n',
    'author': 'Matthew Schinckel',
    'author_email': 'matt@schinckel.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://django-shared-property.readthedocs.io',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>3.8',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
from django.db.models import Expression, F, AutoField
from django.db.models.sql.query import Query
from django.utils.functional import cached_property

from .expressions import ExpressionCol
from .parser import Parser


class SharedPropertyField(AutoField):
    def __init__(self, name, expression, model):
        self.expression = expression
        self.model = model
        super().__init__()
        self.set_attributes_from_name(name)
        self.db_returning = False

    def get_col(self, alias, output_field=None):
        if alias != self.model._meta.db_table or output_field != self:
            return ExpressionCol(
                self.expression,
                self.model,
                alias,
                output_field or self,
            )
        return self.cached_col

    @cached_property
    def cached_col(self):
        return ExpressionCol(
            self.expression,
            self.model,
            self.model._meta.db_table,
            self
        )

    @property
    def output_field(self):
        if getattr(self.expression, 'output_field', None):
            return self.expression.output_field
        return self.expression.resolve_expression(Query(self.model)).output_field

    def contribute_to_class(self, cls, name, **kwargs):
        if cls != self.model:
            # Adding to a concrete subclass.
            field = SharedPropertyField(name, expression=self.expression, model=cls)
            cls._meta.add_field(field, private=True)


class shared_property(object):
    def __init__(self, func):
        if isinstance(func, (Expression, F)):
            self.expression = func
        else:
            self.parsed = Parser(func)
            self.expression = self.parsed.expression
            self.func = func
            context = dict(func.__globals__)
            eval(self.parsed.code, context)
            self.callable = context[func.__code__.co_name]

    def __call__(self, method):
        self.callable = method
        return self

    def __get__(self, instance, cls=None):
        if instance is None:
            return self
        return self.callable(instance)

    def __set__(self, instance, value):
        # We don't really do anything with the value - but we need to
        # override this otherwise instance.refresh_from_db() would stomp
        # over our values.
        if self.callable.__name__ in instance.__dict__ and self.callable(instance) != value:
            raise ValueError('Setting a value that does not match the calculated value is unsupported')
        # However, to prevent an issue where it thinks we
        # have deferred_fields, we want to also store the
        # value on the instance.
        instance.__dict__[self.callable.__name__] = value

    def contribute_to_class(self, cls, name, private_only=False):
        field = SharedPropertyField(name, expression=self.expression, model=cls)
        cls._meta.add_field(field, private=True)
        setattr(cls, field.attname, self)

    def property(self, method):
        return self(method)



from django.db.models.expressions import Expression, F, Q


class ExpressionCol(Expression):
    contains_aggregate = False

    def __init__(self, expression, model, alias=None, target=None):
        self.expression = expression
        self.alias = alias
        self.model = model
        self.target = target

    @property
    def output_field(self):
        return self.target.output_field

    def get_lookup(self, name):
        return self.target.get_lookup(name)

    def get_transform(self, name):
        return self.target.get_transform(name)

    def _resolve_expression(self, query):
        expression = self.target.expression
        if query.model != self.model:
            if isinstance(expression, (F, Q)) and '__' in expression.name:
                resolved = query.__class__(self.model).resolve_ref(expression.name)
                if getattr(resolved, 'model', None) == query.model:
                    return resolved
            parent_alias, _ = query.table_alias(query.model._meta.db_table)
            # We need to work out the path we need to rewrite the F() expression(s) to contain.
            alias, _ = query.table_alias(self.model._meta.db_table)
            while True:
                join = query.alias_map[alias]
                # We need to see which end of this join we need to look up: either name or related_query_name()
                model = join.join_field.model
                if query.table_alias(model._meta.db_table) == join.table_alias:
                    # This is a join that uses a field on the model we are looking at, so
                    # we need the related_query_name
                    expression = retarget_f_expression(expression, join.join_field.related_query_name())
                else:
                    expression = retarget_f_expression(expression, join.join_field.name)
                alias = join.parent_alias
                if alias == parent_alias:
                    break

        # Remove duplicate paths.

        return expression.resolve_expression(query)

    def as_sql(self, compiler, connection):
        resolved = self._resolve_expression(compiler.query)
        alias = getattr(resolved, "alias", None) or self.target.model._meta.db_table
        if alias not in compiler.query.table_map:
            compiler.query.setup_joins(
                self.expression.name.split("__"),
                self.model._meta,
                alias,
            )
        return resolved.as_sql(compiler, connection)

    def get_db_converters(self, connection):
        return self.target.get_db_converters(connection)


def retarget_f_expression(expression, key):
    if isinstance(expression, F):
        return F(f'{key}__{expression.name}')
    if isinstance(expression, Q):
        return Q(**{
            f'{key}__{rest}': value
            for rest, value in expression.children
        })
    expression = expression.copy()
    expression.set_source_expressions([
        retarget_f_expression(expr, key)
        for expr in expression.get_source_expressions()
    ])
    return expression
